/**
 * Teasing Analysis Tool - Updated Version
 * This script processes segmented and processed images to analyze, trace, and manually curate internode regions.
 * Maintained by Pierre Tocquin, University of Liège (ptocquin@uliege.be)
 * Distributed under the MIT License.
 *
 * Description:
 *  - Processes segmented and processed images.
 *  - Identifies internodes based on thresholded particles.
 *  - Interactively validates internodes for inclusion or exclusion.
 *  - Applies morphological transformations (dilate, erode, skeletonize) to refine regions of interest.
 *  - Extracts skeleton traces and computes morphological measurements.
 *  - Outputs curated ROI ZIP and measurement CSV files in the specified folder.
 *
 * Requirements:
 *  - ImageJ version 1.54f or later.
 *
 * Inputs:
 *  - Image pattern: a string pattern used to filter image filenames.
 *  - Working directory: directory that contains the input images (both processed and segmented)
 *    and where the ROI ZIP and CSV measurement files will be saved.
 *
 * Outputs:
 *  - A cleaned ROI ZIP file saved in the specified output folder.
 *  - Measurements data saved as a CSV file in the output folder.
 *
 * Version: 1.2 (2025-02-11)
 */


requires("1.54f"); // Ensure required ImageJ version is available
run("Clear Results"); // Clear any previous results
close("Log"); // Close the log window
setBatchMode(false); // Ensure batch mode is off for user interaction

// Input and output paths provided by the user
#@ String (label="Image pattern: ") pattern // A string pattern used to filter image filenames
#@ File (label="Working directory: ", style="directory") outputFolder // Folder containing images

baseOutputPath = outputFolder + File.separator;

// Get the list of all files in outputFolder
files = getFileList(outputFolder);

for (i = 0; i < files.length; i++) {
    // Only process TIFF files whose names contain the specified pattern
    if (files[i].endsWith(".tif") && files[i].matches(pattern)) {
        imageName = File.getNameWithoutExtension(files[i]); // Extract image name
        
        roiPath = baseOutputPath + imageName + "_roi_cleaned.zip"; // Path for saved ROIs
        resultsPath = baseOutputPath + imageName + "_data.csv"; // Path for saved measurements

        // Define related image filenames
        processed = files[i];
        segmented = processed.replace("resized", "segmented_resized");
        
        // Check if file roiPath exists, if not process skip this file
        if (!File.exists(roiPath)) {
        	print("Processing: "+files[i]);
        	// Open segmented and processed images
        	open(baseOutputPath + segmented);
        	open(baseOutputPath + processed);
	        // Internode analysis ((un)comment to enable/disable if needed)
	        processROI("node", 3, 250, 5000, false);
	        processROI("internode", 2, 10000, 150000, false);
	        processInternodes();
	        checkTraces();
	        if (File.exists(roiPath) && !File.exists(resultsPath)) {
	            // Measurements
	            measureInternodes(130, 40);
	        }
        }
        
        // Close all opened images and results, then reset the ROI manager
        close("*");
        close("Results");
        roiManager("reset");
    }
}

/**
 * Wrapper for process ROI functions.
 */
function processROI(type, group, size_min, size_max, check) {
	// Analyze segmented image
    selectImage(segmented);
    run("Duplicate...", "ignore");
	rename("segmented");
    analyseParticles(group, size_min, size_max);
    close("segmented");

    // Process ROIs in the processed image
    RoiManager.selectGroup(0); // Select the default ROI group
    RoiManager.setGroup(99); // Temporarily set ROIs for review to group 99

    n = roiManager("count"); // Count the number of ROIs
    print("Number of ROI ("+type+") to analyse: " + RoiManager.selected);

	selectImage(processed);
	if (group == 2) {
		roiManager("Show All without labels");
	}
    for (j = 0; j < n; j++) {
        checkROI(j, type, group); // Validate each ROI
    }
    
    selectImage(segmented);
    for (j = 0; j < n; j++) {
        changePixelValue(j, group+10, group, 2); // Validate each ROI
    }

	if (check != true) {
	    // Remove incorrect objects from group 99
	    deleteGroup(99);
	    deleteGroup(group+10);
        renameROI(type, group);
	}
	
    // Save cleaned ROIs to the specified path
    roiManager("save", roiPath);
}

/**
 * Analyze particles in the segmented image.
 * Thresholds the image and identifies particles of specific size.
 */
function analyseParticles(value, min, max) {
	print("Analyse particles "+value);
    setThreshold(value, value, "raw"); // Set raw intensity threshold
    setOption("BlackBackground", true); // Ensure black background for masking
    run("Convert to Mask"); // Convert to binary mask
    run("Analyze Particles...", "size="+min+"-"+max+" add"); // Analyze particles with size constraints
}

/**
 * Check a node or internode ROI for validity with user input.
 * If valid, it is renamed and grouped for keeping; otherwise, it is marked for deletion.
 */
function checkROI(roi, type, group) {
    roiManager("Select", roi); // Select the ROI
    // Process only if not already processsed
    if (Roi.getGroup() == 99) {
    	Roi.setStrokeColor("red"); // Highlight ROI in red
	    Roi.setStrokeWidth(0); // Set stroke width to 0
	    run("Enlarge...", "enlarge=20 pixel"); // Enlarge ROI for review
	    run("To Selection"); // Convert enlarged ROI to selection
	    run("Out [-]"); // Zoom out
	    run("Out [-]"); // Zoom out
	    run("Out [-]"); // Zoom out
	    run("Out [-]"); // Zoom out
	
	    // Prompt user to confirm keeping the roi
	    if (getBoolean("Keep this "+type+"? Clicking NO will remove the object.") == true) {
	    	roiManager("Select", roi); // Select the ROI
	    	roiManager("rename", type + "_" + roi);
	        Roi.setName(type + "_" + roi);
	        Roi.setGroup(group); // Assign to internode group
	    } else {
	    	roiManager("Select", roi); // Select the ROI
	    	roiManager("rename", "_" + type + "_" + roi);
	        Roi.setGroup(group+10); // Mark for deletion
	    }
	}
}

/**
 * Change pixel values in the segmented image
 */
function changePixelValue(roi, group, from, to) {
	roiManager("select", roi); // Select the ROI
	if (Roi.getGroup() == group) {
		print("change "+Roi.getGroup());
		changeValues(from, from, to);
	}
}

/**
 * Delete all ROIs in a specified group.
 */
function deleteGroup(id) {
    RoiManager.selectGroup(id); // Select ROIs in the specified group
    print("Selected for deletion: " + RoiManager.selected); // Log the number of selected ROIs
    if (RoiManager.selected > 0) {
        roiManager("delete"); // Delete selected ROIs
    }
}

/**
 * Rename all internode ROIs sequentially.
 */
function renameROI(type, group) {
    n = roiManager("count"); // Count the number of ROIs
    for (j = 0; j < n; j++) {
        roiManager("select", j); // Select each ROI
        if (Roi.getGroup() == group) {
            roiManager("rename", type+"_" + j); // Rename sequentially
        } 
    }
}

/**
 * Process internodes: apply transformations and trace skeletons.
 */
function processInternodes() {
    l = roiManager("count"); // Count the number of ROIs
	selectImage(processed);
    for (m = 0; m < l; m++) {
    	roiManager("select", m); // Select each ROI
        if (Roi.getGroup() == 2) {
            internodeName = Roi.getName;
            print("Processing internode "+internodeName);
            run("Create Mask");
            selectImage("Mask");

			RoiManager.selectByName(internodeName);
			getSelectionBounds(x, y, width, height);

			run("Crop");
            run("Select None");
            run("Options...", "iterations=10 count=4 black do=Nothing");
			run("Dilate");
			run("Erode");
			run("Erode");            
            run("Skeletonize");
            run("Create Selection");
            getSelectionBounds(xx, yy, wwidth, hheight);
            roiManager("add");
            roiManager("select", RoiManager.size-1);
            roiManager("rename",internodeName+"_skeleton");
            traceInternode();
            getSelectionBounds(xxx, yyy, wwwidth, hhheight);
            
            selectImage(processed);
            RoiManager.selectByName(internodeName);
            setSelectionLocation(x, y);
            RoiManager.selectByName(internodeName+"_skeleton");
		    setSelectionLocation(x+xx, y+yy);
            RoiManager.selectByName(internodeName+"_trace");
		    setSelectionLocation(x+xxx, y+yyy);
		    roiManager("update");
            
            close("Mask");
        } 
    }
    
    // Save cleaned ROIs to the specified path
    roiManager("save", roiPath);
}

function traceInternode() {
    print("Tracing internode "+internodeName);
    Roi.getContainedPoints(xpoints, ypoints);
    //Array.show("coordinates", xpoints, ypoints);
    
    Y_OFFSET = newArray( -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4,  4, 4, 4, 4, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3 );
    X_OFFSET = newArray( -4, -3, -2, -1,  0,  1,  2,  3,  4, 4,  4, 4, 4, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -4, -4, -4, -4, -4, -4 );
    
    STARTX = xpoints[0];
    STARTY = ypoints[0];
    HEADX = xpoints[0];
    HEADY = ypoints[0];
    
    // Looking for the first nearest pixels, evaluating if there is one or two
    nearestPixels = newArray();
    xO = newArray();
    yO = newArray();
    hX = newArray();
    hY = newArray();
    index = newArray();
    for (i = 0; i < X_OFFSET.length; i++) {
        np = getPixel(HEADX+X_OFFSET[i], HEADY+Y_OFFSET[i]);
        if (np == 255) {
            nearestPixels = Array.concat(nearestPixels, np);
            xO = Array.concat(xO, X_OFFSET[i]);
            yO = Array.concat(yO, Y_OFFSET[i]);
            hX = Array.concat(hX, HEADX+X_OFFSET[i]);
            hY = Array.concat(hY, HEADY+Y_OFFSET[i]);
            index = Array.concat(index, i);
        }
    }
    
    Array.sort(nearestPixels, xO, yO, hX, hY, index); // on classe les pixels par index
    //Array.show("starting points before", nearestPixels, xO, yO, hX, hY, index);
    n = nearestPixels.length;
    //print("n ",n);
    
    if(n>0 && n<10) {
        _nearestPixels = newArray();
        _xO = newArray();
        _yO = newArray();
        _hX = newArray();
        _hY = newArray();
        _index = newArray();
        _nearestPixels = Array.concat(_nearestPixels,nearestPixels[0]);
        _xO = Array.concat(_xO,xO[0]);
        _yO = Array.concat(_yO,yO[0]);
        _hX = Array.concat(_hX,hX[0]);
        _hY = Array.concat(_hY,hY[0]);
        _index = Array.concat(_index,index[0]);
        if (n>1) {
            for (i = 1; i < n; i++) {
                if (abs(index[i]-index[i-1]) > 10) {
                    _nearestPixels = Array.concat(_nearestPixels, nearestPixels[i]);
                    _xO = Array.concat(_xO, xO[i]);
                    _yO = Array.concat(_yO, yO[i]);
                    _hX = Array.concat(_hX, hX[i]);
                    _hY = Array.concat(_hY, hY[i]);
                    _index = Array.concat(_index, index[i]);
                }
            }
        }
        
        nearestPixels = _nearestPixels;
        xO = _xO;
        yO = _yO;
        hX = _hX;
        hY = _hY;
        index = _index;
        
        Array.sort(nearestPixels, xO, yO, hX, hY, index); // on classe les pixels les plus sombres en premier
        //Array.show("starting points after", nearestPixels, xO, yO, hX, hY, index);
        //waitForUser;
        
        if (nearestPixels.length < 3) {
            FullTraceX = newArray();
            FullTraceY = newArray();
            for (i = 0; i < nearestPixels.length; i++) {
                //print("i: "+i);
                Y_OFFSET = newArray( -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4,  4, 4, 4, 4, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3 );
                X_OFFSET = newArray( -4, -3, -2, -1,  0,  1,  2,  3,  4, 4,  4, 4, 4, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -4, -4, -4, -4, -4, -4 );
                HEADX = hX[i];
                HEADY = hY[i];
                tX = newArray(STARTX, HEADX); // current trace
                tY = newArray(STARTY, HEADY);
                lXO = xO[i]; // last X offset
                lYO = yO[i];
                shift = index[i];
                WALK = true;
                k = 0;
                while (WALK == true) {
                    WALK = false;
                    Array.rotate(X_OFFSET, 32-shift+3);
                    Array.rotate(Y_OFFSET, 32-shift+3);
                    for (j = 0; j < X_OFFSET.length; j++) {
                        //print("j: "+j);
                        if (abs(X_OFFSET[j]+lXO) > 3 || abs(Y_OFFSET[j]+lYO) >3) { // On évite un 1/2 tour
                            np = getPixel(HEADX+X_OFFSET[j], HEADY+Y_OFFSET[j]);
                            //print("np: "+np);
                            if (np == 255) {
                                lXO = X_OFFSET[j];
                                lYO = Y_OFFSET[j];
                                HEADX = HEADX+lXO;
                                HEADY = HEADY+lYO;
                                tX = Array.concat(tX, HEADX);
                                tY = Array.concat(tY, HEADY);
                                WALK = true;
                                shift = j;
                                j = X_OFFSET.length; // break the loop
                            }
                        }
                    }
                    k = k + 1;
                    if(k == 5000) {
                        WALK = false;
                    }
                }
                if (i == 0) {
                    FullTraceX = Array.concat(FullTraceX, Array.reverse(tX));
                    FullTraceY = Array.concat(FullTraceY, Array.reverse(tY));
                } else {
                    FullTraceX = Array.concat(FullTraceX, tX);
                    FullTraceY = Array.concat(FullTraceY, tY);
                }
            }
            
            //Array.show("Trace", FullTraceX, FullTraceY);
            run("Select None");
            makeSelection("polyline", FullTraceX, FullTraceY);
            roiManager("add");
            RoiManager.select(roiManager("count")-1);
            Roi.setStrokeColor("grey");
            Roi.setStrokeWidth(0.1);
            roiManager("rename",internodeName+"_trace");
        }
    }
}

function checkTraces() {
    l = roiManager("count"); // Count the number of ROIs
	selectImage(processed);
    for (m = 0; m < l; m++) {
        roiManager("select", m);
        s = Roi.getName;
        if(s.endsWith("trace")) {
            run("To Selection");
            run("Out [-]");
            if(getBoolean("Keep this trace ? Clicking NO will remove the object.") == true){
                RoiManager.setGroup(1); // validated traces group
            } else {
                RoiManager.setGroup(99); // to delete
            }
        }
    }
    
    // Save cleaned ROIs to the specified path
    roiManager("save", roiPath);
}

function measureInternodes(threshold, max_th) {
    // Check if roimanager is open, if not open the file roiPath
    if (roiManager("count") == 0) {
        roiManager("Open", roiPath);
    }
	run("Set Measurements...", "area mean standard modal min perimeter feret's display redirect=None decimal=3");
	n = roiManager('count');
	image = newArray();
	labels = newArray();
	feret_diam = newArray();
	perim_diam = newArray();
	length = newArray();
	mean_thickness = newArray();
	std_thickness = newArray();
	for (k = 0; k < n; k++) {
		roiManager('select', k);
		if (Roi.getGroup() == 1) { // measure only validated traces
			image = Array.concat(image,imageName);
		    labels = Array.concat(labels, Roi.getName);
		    Roi.getContainedPoints(FullTraceX, FullTraceY);
			run("Measure");
			//waitForUser;
			feret_diam = Array.concat(feret_diam, getResult("Feret"));
			perim_diam = Array.concat(perim_diam, getResult("Perim.")/2);
			length = Array.concat(length,getResult("Length"));
			run("Clear Results");
			thickness = newArray();
			for (i = 0; i < FullTraceX.length; i = i+5) {
				j = 2;
				while (j < max_th) {
					makeOval(FullTraceX[i]-j/2, FullTraceY[i]-j/2, j, j);
					run("Measure");
					if (getResult("Max") > threshold) {
						diam = j;
						j = max_th; //break
					} else {
						diam = j;
						j = j+2;
					}
				}
				thickness = Array.concat(thickness, diam);
			}
			//Array.show(thickness);
			Array.getStatistics(thickness, min, max, mean, stdDev);
			mean_thickness = Array.concat(mean_thickness, mean);
			std_thickness = Array.concat(std_thickness, stdDev);
		}
	}
	Array.show("Summary", image, labels, feret_diam, perim_diam, length, mean_thickness, std_thickness);
	saveAs("Results", resultsPath);
}
