/*
ImageJ Macro for Node and Internode Annotation

Description:
This macro facilitates the annotation of nodes and internodes in plant images.
It allows users to manually select and modify regions of interest (ROIs) in
segmented images. The macro includes functions for:
- Opening images and associated ROI files.
- Adding or removing nodes and internodes.
- Processing internodes to generate skeletonized traces.
- Renaming and managing ROI annotations.

Author: Pierre Tocquin, Université de Liège
Email: ptocquin@uliege.be

License:
MIT License

Copyright (c) 2024 Pierre Tocquin, Université de Liège
*/

// Add a new node at the cursor position using 'n' key
macro 'Add Node [n]' {
	addNode();
}

// Add a new internode at the cursor position using 'i' key
macro 'Add Internode [i]' {
	addInternode();
}

// Open image and associated files using 'o' key
macro 'Open Images [o]' {
	openImages();
}

// Rename all internodes sequentially using 'r' key
macro 'Rename internodes [r]' {
	renameInternodes();
}

/**
 * openImages: Prompts for a file, opens the original and segmented images,
 * and loads the ROI file if it exists.
 */
function openImages() {
	path = File.openDialog("Select a File");
  	open(path); // open the file
  	dir = File.getParent(path)+"/";
  	file = File.getName(path);
    imageName = File.getNameWithoutExtension(file);
    segmented = file.replace("resized", "segmented_resized");
    roiPath = dir + imageName + "_roi_cleaned.zip";

    // Open the segmented image if it is not
    if (isOpen(segmented) == false) {
        run("Open...", "path=" + segmented);
    }

	// Open the roi file if it exists
	if (File.exists(roiPath)) {
		roiManager("Open", roiPath);
	} else {
		print("No ROI file found at: " + roiPath);
	}
	
	selectImage(file);
	print("Ready!");

}

/**
 * addNode: Adds a new node at the cursor.
 * If the pixel value is 3, asks for confirmation to convert and deletes the old node.
 * Otherwise, duplicates the image, creates a centered rectangular selection with thresholding,
 * then adds, renames, and updates the ROI in the ROI Manager.
 */
function addNode() {
	dir = getInfo("image.directory");
	file = getInfo("image.filename");
    imageName = File.getNameWithoutExtension(file);
    segmented = file.replace("resized", "segmented_resized");
    roiPath = dir + imageName + "_roi_cleaned.zip";
	run("Select None");
	// Get cursor positions
	getCursorLoc(x, y, z, flag);
	// select segmented image
	selectWindow(segmented);
	if(getPixel(x, y) == 3) {
		waitForUser("Convert node to nerve ?");
		deleteNode(x, y);
	} else {
		run("3-3-2 RGB");
		run("Duplicate...", "title=tmp ignore");
		//make a rectangular selection centered at the cursor
		makeRectangle(x-15, y-15, 30, 30);
		// zoom in
		run("To Selection");
		getSelectionBounds(xx, yy, wwidth, hheight);
		if (selectionType() != -1) {
			run("Crop");
			setThreshold(2, 2);
			run("Create Selection");
			getSelectionBounds(xxx, yyy, wwwidth, hhheight);
			selectImage(segmented);
			run("Restore Selection");
			setSelectionLocation(xx+xxx, yy+yyy);
			changeValues(2, 2, 3);
			roiManager("Add");
			n = roiManager("count");
			roiManager("select", n-1);
			roiManager("Rename", "node_xxx");
			renameNodes();
			close("tmp");
		}
	}
	selectImage(file);
	roiManager("Show None");
	roiManager("Show All without labels");
}

/*
 * addInternode(): Create or delete internode ROIs based on the cursor position within an image.
 * If the cursor is within an existing ROI, the ROI is deleted.
 */
function addInternode() {
	dir = getInfo("image.directory");
	file = getInfo("image.filename");
    imageName = File.getNameWithoutExtension(file);
    segmented = file.replace("resized", "segmented_resized");
    roiPath = dir + imageName + "_roi_cleaned.zip";
	run("Select None");
	// Get cursor positions
	getCursorLoc(x, y, z, flag);
	print(x + " " + y + " " + z);
	name = isInternode(x, y);
	if(name != "") {
		waitForUser("ROI found ! Delete "+name+" ?");
		deleteInternode(name);
		//renameInternodes();
	} else {
		waitForUser("Create a new internode ?");
		// select segmented image
		selectImage(segmented);
		doWand(x, y);
		// zoom in
		run("To Selection");
		if (selectionType() != -1) {
			roiManager("Add");
			n = roiManager("count");
			roiManager("select", n-1);
			roiManager("Rename", "internode_" + n);
			Roi.setName("internode_" + n);
			Roi.setGroup(2);
			processInternode();
			selectImage(file);
			run("Select None");
			roiManager("Show All without labels");
		}
	}
}

/**
 * Process internode: apply transformations and trace skeleton.
 */
function processInternode() {
    internodeName = Roi.getName;
    print("Processing internode "+internodeName);
    run("Create Mask");
    selectImage("Mask");

	RoiManager.selectByName(internodeName);
	getSelectionBounds(x, y, width, height);

	run("Crop");
	run("Select None");
    run("Options...", "iterations=10 count=4 black do=Nothing");
	run("Dilate");
	run("Erode");
	run("Erode");            
    run("Skeletonize");
    run("Create Selection");
    getSelectionBounds(xx, yy, wwidth, hheight);
    roiManager("add");
    roiManager("select", RoiManager.size-1);
    roiManager("rename",internodeName+"_skeleton");
    traceInternode();
    getSelectionBounds(xxx, yyy, wwwidth, hhheight);
    
    selectImage(file);
    RoiManager.selectByName(internodeName);
    setSelectionLocation(x, y);
    RoiManager.selectByName(internodeName+"_skeleton");
    setSelectionLocation(x+xx, y+yy);
    RoiManager.selectByName(internodeName+"_trace");
    setSelectionLocation(x+xxx, y+yyy);
    roiManager("update");
    
    close("Mask");
    
    // Save cleaned ROIs to the specified path
    //roiManager("save", roiPath);
}

function traceInternode() {
    print("Tracing internode "+internodeName);
    Roi.getContainedPoints(xpoints, ypoints);
    //Array.show("coordinates", xpoints, ypoints);
    
    Y_OFFSET = newArray( -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4,  4, 4, 4, 4, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3 );
    X_OFFSET = newArray( -4, -3, -2, -1,  0,  1,  2,  3,  4, 4,  4, 4, 4, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -4, -4, -4, -4, -4, -4 );
    
    STARTX = xpoints[0];
    STARTY = ypoints[0];
    HEADX = xpoints[0];
    HEADY = ypoints[0];
    
    // Looking for the first nearest pixels, evaluating if there is one or two
    nearestPixels = newArray();
    xO = newArray();
    yO = newArray();
    hX = newArray();
    hY = newArray();
    index = newArray();
    for (i = 0; i < X_OFFSET.length; i++) {
        np = getPixel(HEADX+X_OFFSET[i], HEADY+Y_OFFSET[i]);
        if (np == 255) {
            nearestPixels = Array.concat(nearestPixels, np);
            xO = Array.concat(xO, X_OFFSET[i]);
            yO = Array.concat(yO, Y_OFFSET[i]);
            hX = Array.concat(hX, HEADX+X_OFFSET[i]);
            hY = Array.concat(hY, HEADY+Y_OFFSET[i]);
            index = Array.concat(index, i);
        }
    }
    
    Array.sort(nearestPixels, xO, yO, hX, hY, index); // on classe les pixels par index
    //Array.show("starting points before", nearestPixels, xO, yO, hX, hY, index);
    n = nearestPixels.length;
    //print("n ",n);
    
    if(n>0 && n<10) {
        _nearestPixels = newArray();
        _xO = newArray();
        _yO = newArray();
        _hX = newArray();
        _hY = newArray();
        _index = newArray();
        _nearestPixels = Array.concat(_nearestPixels,nearestPixels[0]);
        _xO = Array.concat(_xO,xO[0]);
        _yO = Array.concat(_yO,yO[0]);
        _hX = Array.concat(_hX,hX[0]);
        _hY = Array.concat(_hY,hY[0]);
        _index = Array.concat(_index,index[0]);
        if (n>1) {
            for (i = 1; i < n; i++) {
                if (abs(index[i]-index[i-1]) > 10) {
                    _nearestPixels = Array.concat(_nearestPixels, nearestPixels[i]);
                    _xO = Array.concat(_xO, xO[i]);
                    _yO = Array.concat(_yO, yO[i]);
                    _hX = Array.concat(_hX, hX[i]);
                    _hY = Array.concat(_hY, hY[i]);
                    _index = Array.concat(_index, index[i]);
                }
            }
        }
        
        nearestPixels = _nearestPixels;
        xO = _xO;
        yO = _yO;
        hX = _hX;
        hY = _hY;
        index = _index;
        
        Array.sort(nearestPixels, xO, yO, hX, hY, index); // on classe les pixels les plus sombres en premier
        //Array.show("starting points after", nearestPixels, xO, yO, hX, hY, index);
        //waitForUser;
        
        if (nearestPixels.length < 3) {
            FullTraceX = newArray();
            FullTraceY = newArray();
            for (i = 0; i < nearestPixels.length; i++) {
                //print("i: "+i);
                Y_OFFSET = newArray( -4, -4, -4, -4, -4, -4, -4, -4, -4, -3, -2, -1, 0, 1, 2, 3, 4,  4, 4, 4, 4, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3 );
                X_OFFSET = newArray( -4, -3, -2, -1,  0,  1,  2,  3,  4, 4,  4, 4, 4, 4, 4, 4, 4, 3, 2, 1, 0, -1, -2, -3, -4, -4, -4, -4, -4, -4, -4, -4 );
                HEADX = hX[i];
                HEADY = hY[i];
                tX = newArray(STARTX, HEADX); // current trace
                tY = newArray(STARTY, HEADY);
                lXO = xO[i]; // last X offset
                lYO = yO[i];
                shift = index[i];
                WALK = true;
                k = 0;
                while (WALK == true) {
                    WALK = false;
                    Array.rotate(X_OFFSET, 32-shift+3);
                    Array.rotate(Y_OFFSET, 32-shift+3);
                    for (j = 0; j < X_OFFSET.length; j++) {
                        //print("j: "+j);
                        if (abs(X_OFFSET[j]+lXO) > 3 || abs(Y_OFFSET[j]+lYO) >3) { // On évite un 1/2 tour
                            np = getPixel(HEADX+X_OFFSET[j], HEADY+Y_OFFSET[j]);
                            //print("np: "+np);
                            if (np == 255) {
                                lXO = X_OFFSET[j];
                                lYO = Y_OFFSET[j];
                                HEADX = HEADX+lXO;
                                HEADY = HEADY+lYO;
                                tX = Array.concat(tX, HEADX);
                                tY = Array.concat(tY, HEADY);
                                WALK = true;
                                shift = j;
                                j = X_OFFSET.length; // break the loop
                            }
                        }
                    }
                    k = k + 1;
                    if(k == 5000) {
                        WALK = false;
                    }
                }
                if (i == 0) {
                    FullTraceX = Array.concat(FullTraceX, Array.reverse(tX));
                    FullTraceY = Array.concat(FullTraceY, Array.reverse(tY));
                } else {
                    FullTraceX = Array.concat(FullTraceX, tX);
                    FullTraceY = Array.concat(FullTraceY, tY);
                }
            }
            
            //Array.show("Trace", FullTraceX, FullTraceY);
            run("Select None");
            makeSelection("polyline", FullTraceX, FullTraceY);
            roiManager("add");
            RoiManager.select(roiManager("count")-1);
            Roi.setStrokeColor("grey");
            Roi.setStrokeWidth(0.1);
            roiManager("rename",internodeName+"_trace");
        }
    }
}

function checkTraces() {
    l = roiManager("count"); // Count the number of ROIs
	selectImage(processed);
    for (m = 0; m < l; m++) {
        roiManager("select", m);
        s = Roi.getName;
        if(s.endsWith("trace")) {
            run("To Selection");
            run("Out [-]");
            if(getBoolean("Keep this trace ? Clicking NO will remove the object.") == true){
                RoiManager.setGroup(1); // validated traces group
            } else {
                RoiManager.setGroup(99); // to delete
            }
        }
    }
    
    // Save cleaned ROIs to the specified path
    //roiManager("save", roiPath);
}

function isInternode(x, y) {
	n = roiManager("count");
	for(j=0;j<n;j++){
		roiManager("Select", j);
		name = Roi.getName;
		if(name.matches("internode_[0-9]{1,4}") && Roi.contains(x, y)){
			return name;
		}
	}
	return "";
}

function deleteNode(x, y) {
	selectImage(segmented);
	n = roiManager("count");
	for(j=0;j<n;j++){
		roiManager("Select", j);
		name = Roi.getName;
		if(Roi.contains(x, y) && name.matches("node_.*")){
			changeValues(3, 3, 2);
			roiManager("delete");
			j = n;
			renameNodes();
		}
	}
}

function renameNodes() {
	l = roiManager("count");
	for (j = 0; j < l; j++) {
		roiManager("select", j);
		name = Roi.getName;
		if(name.matches("node_.*")){
			roiManager("rename", "node_"+j);
			Roi.setGroup(3);
			roiManager("Update");
		}	
	}
}

function deleteInternode(name) {
	print("Delete internode "+name);
	n = roiManager("count");
	for(j=0;j<n;j++){
		roiManager("Select", j);
		fullname = Roi.getName;
		print("fullname: "+fullname);
		if(fullname.matches(name)){
			print(j+" "+n+" "+fullname);
			roiManager("Select", j);
			Roi.getContainedPoints(xpoints, ypoints);
			roiManager("delete");
			j--;
			n--;
			for (k=0; k<n; k++) {
				roiManager("Select", k);
				Roi.getContainedPoints(x, y);
				for (v=0; v < xpoints.length; v++) {
					if (xpoints[v] == x[0] && ypoints[v] == y[0]){
						subname = Roi.getName;
						print(subname+" Trouvé !");
						roiManager("delete");
						k--;
						n--;
					}
				}
			}
		}
	}
	print(name + "deleted"
	run("Select None");
}

function renameInternodes() {
	print("rename internodes");
	l = roiManager("count");
	for (j = 0; j < l; j++) {
		roiManager("select", j);
		fullname = Roi.getName;
		print(fullname);
		if(fullname.matches("internode_[0-9]+.*")){
			print(fullname+" becomes _"+fullname);
			roiManager("rename", "_"+fullname);
			roiManager("Update");
		}
	}
	for (j = 0; j < l; j++) {
		roiManager("select", j);
		fullname = Roi.getName;
		print("second loop: "+fullname);
		if(fullname.matches("_internode_[0-9]+")){
			Roi.getContainedPoints(xpoints, ypoints);
			print(fullname+" becomes internode_"+j);
			roiManager("rename", "internode_"+j);
			Roi.setGroup(2);
			roiManager("Update");
			for(k=0;k<l;k++){
				roiManager("Select", k);
				Roi.getContainedPoints(x, y);
				for (v=0; v < xpoints.length; v++) {
					if (xpoints[v] == x[0] && ypoints[v] == y[0]){
						name = Roi.getName;
						if(name.endsWith("skeleton")){
							print(fullname+" becomes internode_"+j+"_skeleton");
							roiManager("rename", "internode_"+j+"_skeleton");
							Roi.setGroup(2);
							roiManager("Update");
						}
						if(name.endsWith("trace")){
							print(fullname+" becomes internode_"+j+"_trace");
							roiManager("rename", "internode_"+j+"_trace");
							Roi.setGroup(2);
							roiManager("Update");
						}
					}
				}	
			}
		}
	}
}

