"""
Author: Pierre Tocquin (University of Liège, ptocquin@uliege.be)
Description: This script automates the processing of CZI images by converting them to TIFF,
             analyzing them with Ilastik, and resizing the input and segmented outputs.
                The pipeline consists of the following steps:
                1. Assemble mosaic from CZI tiles into a single TIFF file.
                2. Process the mosaic with Ilastik for segmentation.
                3. Correct segmentation issues in the segmented image:
                    - Subsample the image if it is too large to process in memory.
                    - Preserve true nodes of Ranvier by filtering red clusters based on size and aspect ratio.
                    - Apply morphological operations to smooth and refine segmentation results.
             
             The script uses a YAML configuration file to define global parameters such as:
             - Output directory for processed files (`output_dir`)
             - Compression type for TIFF files (`compression`)
             - Resize factor for scaling images (`resize_factor`)
             - Path to the Ilastik project file (`project`)
             - Path to the Ilastik executable (`ilastik`)
             - Minimum size of red clusters to preserve as true nodes of Ranvier (`area_threshold`)
             - Minimum and maximum aspect ratios to identify true nodes of Ranvier (`aspect_ratio_min`, `aspect_ratio_max`)

Usage: python process_pipeline.py file1.czi file2.czi ... --config /path/to/config.yaml

Dependencies:
    - PyYAML: For reading the YAML configuration file.
    - aicspylibczi: For processing CZI files.
    - tifffile: For handling TIFF files.
    - OpenCV: For image resizing and morphological operations.
    - Ilastik: For segmentation in headless mode.

License: MIT License
Copyright (c) 2024 Pierre Tocquin
"""

import argparse
import os
import cv2
import subprocess
import yaml
import tifffile as tiff
import numpy as np
import psutil
import logging
from aicspylibczi import CziFile


def configure_logging(log_level="INFO"):
    """
    Configures the logging system to include timestamps and log levels.

    Args:
        log_level (str): Logging level as a string (e.g., DEBUG, INFO, WARNING).
    """
    numeric_level = getattr(logging, log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f"Invalid log level: {log_level}")

    logging.basicConfig(
        level=numeric_level,
        format='%(asctime)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    logging.info(f"Logging configured at {log_level} level")

def load_config(config_path):
    """
    Loads and validates the configuration from a YAML file.

    Args:
        config_path (str): Path to the YAML configuration file.

    Returns:
        dict: Dictionary of configuration options.
    """
    with open(config_path, 'r') as file:
        config = yaml.safe_load(file)
    
    required_keys = [
        'output_dir', 'compression', 'resize_factor', 
        'project', 'ilastik', 'area_threshold', 
        'aspect_ratio_min', 'aspect_ratio_max', 'segment_size',
        'max_image_size'
    ]    
    for key in required_keys:
        if key not in config:
            raise ValueError(f"Missing required configuration key: {key}")
    
    return config

def log_memory_usage(step_name):
    """
    Logs the memory usage at a given point in the script.
    """
    process = psutil.Process(os.getpid())
    memory_mb = process.memory_info().rss / (1024 ** 2)
    logging.debug(f"[Memory Usage] After {step_name}: {memory_mb:.2f} MB")

def process_czi_file(file_path, config):
    """
    Processes a CZI file, assembles mosaics, analyzes them with Ilastik, and corrects segmentation issues.
    Each series within the CZI file is processed sequentially.

    Workflow:
        1. Assemble mosaic from CZI tiles into a single TIFF file.
        2. Process the mosaic with Ilastik for segmentation.
        3. Correct segmentation issues in the segmented image:
            - Subsample the image if it is too large to process in memory.
            - Preserve true nodes of Ranvier by filtering red clusters based on size and aspect ratio.
            - Apply morphological operations to smooth and refine segmentation results.
        4. Resize both the original mosaic and segmented outputs.
        5. Remove intermediate files to optimize disk usage.

    Args:
        file_path (str): Path to the CZI file to process.
        config (dict): Configuration options loaded from YAML, including:
            - output_dir (str): Directory for saving processed files.
            - compression (str): Compression type for TIFF files (e.g., "lzw").
            - resize_factor (float): Scaling factor for resizing images.
            - project (str): Path to the Ilastik project file.
            - ilastik (str): Path to the Ilastik executable.
            - area_threshold (int): Minimum size for preserving red clusters as nodes of Ranvier.
            - closing_kernel_size (int): Kernel size for filling interruptions in blue regions.
            - aspect_ratio_min (float): Minimum aspect ratio for identifying nodes of Ranvier.
            - aspect_ratio_max (float): Maximum aspect ratio for identifying nodes of Ranvier.
    """

    print(f"Processing file: {file_path}")

    # Load the CZI file
    czi = CziFile(file_path)
    dims = czi.dims  # Example: "HSCMYXA"
    x_idx = dims.index('X')
    y_idx = dims.index('Y')

    # Extract the file name without the extension
    input_filename = os.path.splitext(os.path.basename(file_path))[0]

    num_series = len(czi.get_dims_shape())
    logging.info(f"Number of detected series (S): {num_series}")

    for s in range(num_series):
        logging.info(f"  Processing series {s}")

        # Retrieve information about the tiles for this series
        tiles_nb = len(czi.get_all_mosaic_tile_bounding_boxes(S=s, C=0, A=0, H=0))
        logging.info(f"    Number of tiles for S={s}: {tiles_nb}")

        # Initialize the global dimensions of the mosaic
        max_x, max_y = 0, 0
        tiles = []

        # Iterate through the tiles and retrieve their data
        for m in range(tiles_nb):
            # Read the tile
            tile, _ = czi.read_image(S=s, C=0, M=m, A=0, H=0)

            # Retrieve the tile's coordinates and dimensions
            tile_boundings = czi.get_mosaic_tile_bounding_box(S=s, C=0, M=m, A=0, H=0)

            # Use the first tile's coordinates to adjust the offsets
            if m == 0:
                shift_x = tile_boundings.x
                shift_y = tile_boundings.y

            # Compute the corrected positions
            pos_x = tile_boundings.x - shift_x
            pos_y = tile_boundings.y - shift_y
            width, height = tile_boundings.w, tile_boundings.h

            # Add the tile data to the list
            tiles.append((tile, pos_x, pos_y, width, height))

            # Update the global dimensions
            max_x = max(max_x, pos_x + width)
            max_y = max(max_y, pos_y + height)

        # Create an empty image for the final mosaic
        mosaic_image = np.zeros((max_y, max_x), dtype=tiles[0][0].dtype)

        # Assemble the tiles into the final image
        for tile, pos_x, pos_y, width, height in tiles:
            slices = [slice(None)] * len(dims)
            slices[x_idx] = slice(0, width)
            slices[y_idx] = slice(0, height)
            mosaic_image[pos_y:pos_y + height, pos_x:pos_x + width] = tile[0, 0, 0, 0, :, :, 1]

        # Save the assembled mosaic as a TIFF with a series suffix
        output_filename = os.path.join(config['output_dir'], f"{input_filename}_series_{s}.tif")
        tiff.imwrite(output_filename, mosaic_image, compression=config['compression'])

        logging.info(f"Generated TIFF for series {s}: {output_filename}")

        # Process the generated TIFF with Ilastik
        segmented_path = run_ilastik(output_filename, config['project'], config['ilastik'])

        # Load segmented image for correction
        segmented_image = tiff.imread(segmented_path)

        # If segmented image is too big, subsample it, process each subsample, and reassemble the final image
        if segmented_image.size > config['max_image_size']:
            logging.warning(f"Segmented image too large. Subsampling for processing.")
            segment_size = config['segment_size']
            segmented_subsampled = np.zeros(segmented_image.shape, dtype=segmented_image.dtype)
            for y in range(0, segmented_image.shape[0], segment_size):
                for x in range(0, segmented_image.shape[1], segment_size):
                    logging.info(f"Processing segment at ({y}, {x})")
                    segment = segmented_image[y:y + segment_size, x:x + segment_size]
                    corrected_segment = correct_segmentation_issues(
                        segment,
                        area_threshold=config['area_threshold'],
                        aspect_ratio_min=config['aspect_ratio_min'],
                        aspect_ratio_max=config['aspect_ratio_max']
                    )
                    segmented_subsampled[y:y + segment_size, x:x + segment_size] = corrected_segment

            # Save the subsampled segmented image
            tiff.imwrite(segmented_path, segmented_subsampled.astype(np.uint8))
            logging.info(f"Subsampled segmented image saved: {segmented_path}")
        
        else:  
            # Correct segmentation problems (red clusters and interruptions)
            corrected_image = correct_segmentation_issues(
                segmented_image,
                area_threshold=config['area_threshold'],
                aspect_ratio_min=config['aspect_ratio_min'],
                aspect_ratio_max=config['aspect_ratio_max']
            )

            # Save corrected image
            tiff.imwrite(segmented_path, corrected_image.astype(np.uint8))
            logging.info(f"Corrected segmentation issues in segmented image: {segmented_path}")

        # Save Images and delete intermediate files
        save_image(output_filename)
        save_image(segmented_path)

def run_ilastik(input_path, project_path, ilastik_path, output_format="tif", export_source="Simple Segmentation"):
    """
    Runs Ilastik in headless mode on the given input file.
    """
    output_filename_format = os.path.join(
        os.path.dirname(input_path), f"{os.path.splitext(os.path.basename(input_path))[0]}_segmented"
    )
    cmd = [
        ilastik_path,
        "--headless",
        f"--project={project_path}",
        f"--output_format={output_format}",
        f"--export_source={export_source}",
        f"--output_filename_format={output_filename_format}",
        f"--raw-data={input_path}"
    ]
    try:
        subprocess.run(cmd, check=True)
        segmented_path = f"{output_filename_format}.{output_format}"
        print(f"Processed: {input_path} -> {segmented_path}")
        return segmented_path
    except subprocess.CalledProcessError as e:
        print(f"Error processing: {input_path}")
        raise e

def save_image(image_path, compression='lzw'):

    image = tiff.imread(image_path)
    if image is None:
        raise FileNotFoundError(f"Could not read image: {image_path}")
    
    # Save image
    file_dir, file_name = os.path.split(image_path)
    file_base, file_ext = os.path.splitext(file_name)
    resized_path = os.path.join(file_dir, f"{file_base}_resized{file_ext}")
    # If image is too large (> 2.000.000.0000 pixels), split it in 2 parts (horizontally splitted)
    if image.size > 2000000000:
        logging.info(f"Image too large, splitting in 2 parts")
        half_width = int(image.shape[1] / 2)
        image1 = image[:, :half_width]
        image2 = image[:, half_width:]
        resized_path1 = os.path.join(file_dir, f"{file_base}_resized1{file_ext}")
        resized_path2 = os.path.join(file_dir, f"{file_base}_resized2{file_ext}")
        tiff.imwrite(resized_path1, image1, compression=compression)
        tiff.imwrite(resized_path2, image2, compression=compression)
        logging.info(f"Saved image: {resized_path1}")
        logging.info(f"Saved image: {resized_path2}")
        return
    else:
        tiff.imwrite(resized_path, image, compression=compression)
        logging.info(f"Saved image: {resized_path}")

    # Remove the original image
    os.remove(image_path)

def resize_image(image_path, scale_factor, interpolation, patch_size=1024):
    """
    Resizes an image by processing it in smaller patches to avoid memory overflow.

    Args:
        image_path (str): Path to the image to resize.
        scale_factor (float): Scaling factor for resizing.
        interpolation (int): OpenCV interpolation method.
        patch_size (int): Size of patches to process (default: 1024).

    Returns:
        str: Path to the resized image.
    """
    image = tiff.imread(image_path)
    if image is None:
        raise FileNotFoundError(f"Could not read image: {image_path}")

    # Calculate new size
    new_height = int(image.shape[0] * scale_factor)
    new_width = int(image.shape[1] * scale_factor)
    resized_image = np.zeros((new_height, new_width), dtype=image.dtype)

    # Process image in patches
    for y in range(0, image.shape[0], patch_size):
        for x in range(0, image.shape[1], patch_size):
            # Extract patch
            patch = image[y:y + patch_size, x:x + patch_size]

            # Calculate the new size for the patch
            patch_new_height = int(patch.shape[0] * scale_factor)
            patch_new_width = int(patch.shape[1] * scale_factor)

            # Resize patch
            resized_patch = cv2.resize(patch, (patch_new_width, patch_new_height), interpolation=interpolation)

            # Determine position in resized image
            start_y = int(y * scale_factor)
            start_x = int(x * scale_factor)
            end_y = start_y + resized_patch.shape[0]
            end_x = start_x + resized_patch.shape[1]

            # Place resized patch in the output image
            resized_image[start_y:end_y, start_x:end_x] = resized_patch

    # Save resized image
    file_dir, file_name = os.path.split(image_path)
    file_base, file_ext = os.path.splitext(file_name)
    resized_path = os.path.join(file_dir, f"{file_base}_resized{file_ext}")
    tiff.imwrite(resized_path, resized_image)
    print(f"Saved resized image: {resized_path}")
    return resized_path

def correct_segmentation_issues(image, area_threshold=750, aspect_ratio_min=1, aspect_ratio_max=5):
    """
    Corrects segmentation issues in the processed image:
    1. Preserves large red clusters (true nodes of Ranvier) based on area and shape.

    Args:
        image (numpy.ndarray): The segmented image where white (1), blue (2), and red (3) represent classes.
        area_threshold (int): Minimum area to preserve a red cluster as a true node of Ranvier.
        closing_kernel_size (int): Size of the morphological closing kernel (must be odd).
        aspect_ratio_min (float): Minimum aspect ratio to preserve a cluster as a true node of Ranvier.
        aspect_ratio_max (float): Maximum aspect ratio to preserve a cluster as a true node of Ranvier.

    Returns:
        numpy.ndarray: Corrected segmented image.
    """

    logging.info(f"Start segmentation issues correction")

    logging.debug(f"    ...Apply blur filter")
    # Step 1: Apply a median filter to smooth the final image
    image = cv2.medianBlur(image, 7)

    # Step 2: Preserve large red clusters (true nodes of Ranvier)
    red_mask = (image == 3).astype(np.uint8)
    num_labels, labels, stats, _ = cv2.connectedComponentsWithStatsWithAlgorithm(red_mask, connectivity=8, ltype=cv2.CV_32S, ccltype=cv2.CCL_GRANA)

    logging.info(f"    ...Evaluate {num_labels} labels")
    for label in range(1, num_labels):
        # Extract the mask for the current cluster
        cluster_mask = (labels == label).astype(np.uint8)

        logging.debug(f"        ...Evaluate label {label}")
        area = stats[label, cv2.CC_STAT_AREA]
        width = stats[label, cv2.CC_STAT_WIDTH]
        height = stats[label, cv2.CC_STAT_HEIGHT]
        # Calculate aspect ratio
        aspect_ratio = max(width, height) / min(width, height)
        logging.debug(f"        ... w: {width}, h: {height}, ratio: {aspect_ratio}, area: {area}")

        # If the cluster is on the edge of the image, skip it
        if np.any(cluster_mask[0, :]) or np.any(cluster_mask[:, 0]) or np.any(cluster_mask[-1, :]) or np.any(cluster_mask[:, -1]):
            logging.debug(f"            ...edge cluster")
            continue

        # Skip invalid clusters
        if width == 0 or height == 0:
            logging.debug(f"            ...invalid")
            continue  
        
        if area > area_threshold:
            logging.debug(f"            ...good")
            if aspect_ratio <= aspect_ratio_min or aspect_ratio >= aspect_ratio_max:
                logging.debug(f"            ...but aspect ratio KO")
                # Modify the pixels included in the outline so that they have the value 4
                logging.debug(f"            ...converting to nerve 4")
                image[cluster_mask == 1] = 4
            else: 
                logging.debug(f"            ...aspect ratio OK... keep as is")
        else:
            logging.debug(f"            ...too small, converting to nerve")
            image[cluster_mask == 1] = 2

    logging.info(f"    ...Done !")
    return image

def main():
    parser = argparse.ArgumentParser(description="Automate CZI to Ilastik processing pipeline.")
    parser.add_argument("files", nargs='+', help="List of CZI files to process.")
    parser.add_argument("--config", required=True, help="Path to the YAML configuration file.")
    args = parser.parse_args()

    # Load configuration
    config = load_config(args.config)

    # Configure logging based on the config file
    log_level = config.get("log_level", "INFO")  # Default to INFO if not specified
    configure_logging(log_level)

    # Ensure output directory exists
    if not os.path.isdir(config['output_dir']):
        os.makedirs(config['output_dir'])

    # Process each file through the pipeline
    for file_path in args.files:
        if not os.path.isfile(file_path):
            logging.warning(f"File not found: {file_path}")
            continue
        process_czi_file(file_path, config)

    logging.info("Processing complete") 

if __name__ == "__main__":
    main()
