"""
Author: Pierre Tocquin (University of Liège, ptocquin@uliege.be)
Description: This script assembles mosaics from CZI files provided via the command line.
             It dynamically identifies the relevant dimensions (X, Y, A) and generates
             losslessly compressed TIFF files, with optional resizing.

Usage: python pre-teasing.py file1.czi file2.czi ... [--output-dir /output/path] [--compression deflate/lzw] [--resize-factor 0.25]

License: MIT License
Copyright (c) 2024 Pierre Tocquin

Last updated: 2024-11-29
"""

import argparse
from aicspylibczi import CziFile
import numpy as np
import tifffile as tiff
import os
import cv2

def process_czi_file(file_path, output_dir, compression, resize_factor):
    """
    Processes a CZI file, assembles mosaics, optionally resizes them, and saves as compressed TIFF.

    Args:
        file_path (str): Path to the CZI file to process.
        output_dir (str): Directory where results will be saved.
        compression (str): Type of TIFF compression to use.
        resize_factor (float): Factor to resize the image (1.0 = original size, 0.25 = 25%).
    """
    print(f"Processing file: {file_path}")

    # Load the CZI file
    czi = CziFile(file_path)

    # Identify available dimensions and their positions
    dims = czi.dims  # Example: "HSCMYXA"
    x_idx = dims.index('X')
    y_idx = dims.index('Y')

    # Extract the file name without the extension
    input_filename = os.path.splitext(os.path.basename(file_path))[0]

    # Identify available dimensions
    dims_info = czi.get_dims_shape()
    num_series = len(dims_info)  # Each block in dims_info corresponds to a series

    print(f"Number of detected series (S): {num_series}")

    # Loop through each series
    for s in range(num_series):
        print(f"  Processing series {s}")

        # Retrieve information about the tiles for this series
        tiles_nb = len(czi.get_all_mosaic_tile_bounding_boxes(S=s, C=0, A=0, H=0))
        print(f"    Number of tiles for S={s}: {tiles_nb}")

        # Initialize the global dimensions of the mosaic
        max_x, max_y = 0, 0
        tiles = []

        # Iterate through the tiles and retrieve their data
        for m in range(tiles_nb):
            # Read the tile
            tile, _ = czi.read_image(S=s, C=0, M=m, A=0, H=0)

            # Retrieve the tile's coordinates and dimensions
            tile_boundings = czi.get_mosaic_tile_bounding_box(S=s, C=0, M=m, A=0, H=0)

            # Use the first tile's coordinates to adjust the offsets
            if m == 0:
                shift_x = tile_boundings.x
                shift_y = tile_boundings.y

            # Compute the corrected positions
            pos_x = tile_boundings.x - shift_x
            pos_y = tile_boundings.y - shift_y
            width, height = tile_boundings.w, tile_boundings.h

            # Add the tile data to the list
            tiles.append((tile, pos_x, pos_y, width, height))

            # Update the global dimensions
            max_x = max(max_x, pos_x + width)
            max_y = max(max_y, pos_y + height)

        # Create an empty image for the final mosaic
        mosaic_image = np.zeros((max_y, max_x), dtype=tiles[0][0].dtype)

        # Assemble the tiles into the final image
        for tile, pos_x, pos_y, width, height in tiles:
            slices = [slice(None)] * len(dims)
            slices[x_idx] = slice(0, width)
            slices[y_idx] = slice(0, height)
            mosaic_image[pos_y:pos_y + height, pos_x:pos_x + width] = tile[0, 0, 0, 0, :, :, 1]

        # Resize the mosaic if a resize factor is specified
        if resize_factor != 1.0:
            print(f"    Resizing mosaic to {resize_factor * 100:.1f}% of the original size")
            mosaic_image = resize_with_opencv(mosaic_image, resize_factor)

        # Save the assembled mosaic as a TIFF with a series suffix
        output_filename = os.path.splitext(os.path.basename(file_path))[0]
        output_path = os.path.join(output_dir, f"{output_filename}_series_{s}_processed.tif")
        tiff.imwrite(output_path, mosaic_image, compression=compression)
        print(f"    Mosaic for S={s} saved: {output_path}")

def resize_with_opencv(image, resize_factor):
    """
    Resizes an image using OpenCV.

    Args:
        image (ndarray): The input image array to resize.
        resize_factor (float): The resize factor (e.g., 0.25 for 25% size).

    Returns:
        ndarray: The resized image.
    """
    new_size = (int(image.shape[1] * resize_factor), int(image.shape[0] * resize_factor))
    return cv2.resize(image, new_size, interpolation=cv2.INTER_AREA)


def main():
    # Configure command-line arguments
    parser = argparse.ArgumentParser(description="Assemble mosaics from CZI files.")
    parser.add_argument(
        "files",
        nargs='+',
        help="List of CZI files to process (example: file1.czi file2.czi ...)"
    )
    parser.add_argument(
        "--output-dir",
        default=".",
        help="Directory where output files will be saved (default: current directory)."
    )
    parser.add_argument(
        "--compression",
        default="none",
        choices=["none", "deflate", "lzw"],
        help="Type of compression for TIFF files (default: none)."
    )
    parser.add_argument(
        "--resize-factor",
        type=float,
        default=1.0,
        help="Factor to resize the image (default: 1.0 = no resizing, e.g., 0.25 = 25%)."
    )
    args = parser.parse_args()

    # Check if the output directory exists
    if not os.path.isdir(args.output_dir):
        print(f"Error: The specified output directory does not exist: {args.output_dir}")
        exit(1)
        
    # Process each file provided as an argument
    for file_path in args.files:
        if not os.path.isfile(file_path):
            print(f"File not found: {file_path}")
            continue
        process_czi_file(file_path, args.output_dir, args.compression, args.resize_factor)

if __name__ == "__main__":
    main()
