"""
Author: Pierre Tocquin (University of Liège, ptocquin@uliege.be)
Description: This script processes image files using Ilastik in headless mode and resizes the input and output images.
             Input images are resized with quality-preserving interpolation, and segmented outputs use nearest-neighbor interpolation to preserve segmentation.

Usage: python batch_ilastik.py file1.tif file2.tif ... [--project teasing.ilp] [--ilastik /path/to/ilastik] [--scale-factor 0.5]

License: MIT License
Copyright (c) 2024 Pierre Tocquin

Last updated: 2024-12-02
"""

import os
import cv2
import subprocess
import argparse
import tifffile as tiff

def resize_image(image_path, scale_factor, interpolation):
    """
    Resizes an image using OpenCV.

    Args:
        image_path (str): Path to the image file to resize.
        scale_factor (float): Scale factor for resizing (e.g., 0.5 = 50%).
        interpolation (int): Interpolation method for resizing.

    Returns:
        str: Path to the resized image file.
    """
    image = tiff.imread(image_path)
    if image is None:
        raise FileNotFoundError(f"Could not read image: {image_path}")
    
    new_size = (int(image.shape[1] * scale_factor), int(image.shape[0] * scale_factor))
    resized_image = cv2.resize(image, new_size, interpolation=interpolation)

    # Generate resized file name with '_resized' suffix
    file_dir, file_name = os.path.split(image_path)
    file_base, file_ext = os.path.splitext(file_name)
    resized_path = os.path.join(file_dir, f"{file_base}_resized{file_ext}")
    
    cv2.imwrite(resized_path, resized_image)
    print(f"Saved resized image: {resized_path}")
    return resized_path

def run_ilastik(input_path, project_path, ilastik_path, output_format, export_source):
    """
    Runs Ilastik in headless mode on the given input file.

    Args:
        input_path (str): Path to the input image file.
        project_path (str): Path to the Ilastik project file.
        ilastik_path (str): Path to the Ilastik executable.
        output_format (str): Format for the output file (e.g., "tif").
        export_source (str): Export source for Ilastik (e.g., "Simple Segmentation").

    Returns:
        str: Path to the segmented output file.
    """
    output_filename_format = f"{os.path.dirname(input_path)}/{os.path.splitext(os.path.basename(input_path))[0]}_segmented"
    cmd = [
        ilastik_path,
        "--headless",
        f"--project={project_path}",
        f"--output_format={output_format}",
        f"--export_source={export_source}",
        f"--output_filename_format={output_filename_format}",
        f"--raw-data={input_path}"
    ]
    try:
        subprocess.run(cmd, check=True)
        segmented_path = f"{output_filename_format}.{output_format}"
        print(f"Processed: {input_path} -> {segmented_path}")
        return segmented_path
    except subprocess.CalledProcessError as e:
        print(f"Error processing: {input_path}")
        raise e

def process_images(files, project_path, ilastik_path, scale_factor):
    """
    Processes input files through Ilastik and resizes input and output images.

    Args:
        files (list): List of input image file paths.
        project_path (str): Path to the Ilastik project file.
        ilastik_path (str): Path to the Ilastik executable.
        scale_factor (float): Scale factor for resizing images.
    """
    output_format = "tif"
    export_source = "Simple Segmentation"

    for file_path in files:
        if not os.path.isfile(file_path):
            print(f"File not found: {file_path}")
            continue
        
        # Run Ilastik
        try:
            segmented_path = run_ilastik(file_path, project_path, ilastik_path, output_format, export_source)
        except Exception as e:
            print(f"Error during Ilastik processing for {file_path}: {e}")
            continue

        # Resize input image
        # INTER_AREA is used for resizing the input image when reducing its size.
        # This method computes the pixel value as an average of the neighboring pixels,
        # which helps to preserve details while reducing noise and avoiding aliasing.
        # It is particularly suitable for microscopy images, where maintaining high-quality
        # detail is essential during downscaling.
        resize_image(file_path, scale_factor, cv2.INTER_AREA)
        
        # Resize segmented output
        # INTER_NEAREST_EXACT is used for resizing the segmented output image.
        # This method ensures an exact nearest-neighbor interpolation, avoiding
        # rounding errors that can occur with INTER_NEAREST.
        # It is critical for preserving the discrete class values in segmentation masks
        # (e.g., 0, 1, 2, 3) without introducing any unintended intermediate values.
        # This is especially important in segmentation tasks where precision and label integrity are required.
        resize_image(segmented_path, scale_factor, cv2.INTER_NEAREST_EXACT)

def main():
    """
    Main function for parsing arguments and processing files.
    """
    parser = argparse.ArgumentParser(description="Batch process images with Ilastik and resize them.")
    parser.add_argument(
        "files", 
        nargs='+', 
        help="List of image files to process."
    )
    parser.add_argument(
        "--project", 
        required=True, 
        help="Path to the Ilastik project file (.ilp)."
    )
    parser.add_argument(
        "--ilastik", 
        required=True, 
        help="Path to the Ilastik executable."
    )
    parser.add_argument(
        "--scale-factor", 
        type=float, 
        required=True, 
        help="Scale factor for resizing (e.g., 0.5 for 50%)."
    )
    args = parser.parse_args()

    try:
        process_images(args.files, args.project, args.ilastik, args.scale_factor)
    except Exception as e:
        print(f"An error occurred: {e}")

if __name__ == "__main__":
    main()
